
/* 

 cpd-11-755-2015 =  http://www.clim-past-discuss.net/11/755/2015/cpd-11-755-2015.pdf 
( Optimal Ranking Regime analysis of TreeFlow dendrohydrological reconstructions)


 calc_orr.c Compiles on GNU g++ compiler via:  g++   calc_orr.c   -o calc_orr
   Or, if the included file makefile is in this program's directoru
   enter 'make calc_orr' at the prompt 


Inputs:
data.in - input data series consisting of year, data pairs (see CALC_ORR.scr)
stnindx - stnindx determining the y coordinate of corresponding Z series plots

Outputs:
calc_orr.dat : optimal ranking regimes of data.in input series

            Col1  =   stnindx 
            Col2  =   Begin year of Regime
            Col3  =   End year of Regime
            Col4  =   Mann Whitney Z stat of Regime
            Col5  =   abs(Mann Whitney Z stat of Regime)

calc_orr.dat.all : all significant ranking regimes (@ 90% significance)  of data.in input series. 
Col1-5 same as calc_orr.dat

series.dat : data.in series + filtered or detrended series

*/


#include <stdio.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>  

#define  mxyrs 1300 /* The highest possible number of years in your input time series  */
#define  shrtwin 6   /* shortest window size */
#define  longwin 100  /* longest window size  */
#define  nnsamp longwin-shrtwin+1 /* The number of sampling windows tested, shrtwin to longwin */
#define  maxlag 3    /* maximum lag for calculation of AR parameters */
#define  nr 50000    /* The number of null U stats generated in the Monte Carlo simulations */ 
#define  nmax 130000 /* The maximum number of significant (Z > 1.645) Z stats in each time series
                      this value may need to be increased if mxyrs and/or nnsamp  is increased */ 
#define  bufsize 128
/*--------------------------FUNCTION PROTOTYPES---------------------------------*/

void bubble1(double absig[mxyrs], int index[mxyrs], int* nsig);

void rank(float nmodel[mxyrs], float rnmodel[mxyrs], int nyrs);

void calc_runmean(float aflo[mxyrs], float runmean[mxyrs], int nyrs, int tsamp);

float rabs(float,float);

float array_min(float avmwu[nnsamp]);

float array_max(float avmwu[nnsamp]);

void detrend(float aflo[mxyrs], float daflo[mxyrs], int nyrs, float* dslope, float* yint, float* slpct);

void gettrend(float aflo[mxyrs], int nyrs, float* dslope);

void lanfltr(float aflow[mxyrs], float lpflow[mxyrs], float hpflow[mxyrs], int nyrs, float* tc1);

void getacor(float daflow[mxyrs], float acor[maxlag], int nyrs, float* flmean, float* flvrnc);

void getar(float acor[maxlag], float aic[maxlag], float phibest[maxlag], int nyrs,
           float errvr[maxlag],float flmean, float flvrnc, int* ibest, float* z);           

void getmarkov3(float nmodel[mxyrs], int nyrs, float* mmean, float* mvrnc, float flmean, 
                float c, float* phi1, float* phi2, float* phi3); 

/*-----------------------------MAIN PROGRAM--------------------------------------*/

int main ()
{

  char buf[bufsize],cbuf[20],stnid[3];
  char model_name[20];

  int ibest,ic,nwrote,pass;
  int nyrs, iyr, year, ibegyr, iendyr, iyre, iyrb;
  int nsamp[nnsamp];
  int i,j,k,l,m,n,n1,n2,n3;
  int yr1[nmax+1],yr2[nmax+1];
  int yrbeg[nmax+1],yrend[nmax+1],index[nmax+1];
  int nsig,no[nnsamp]={0},ireps[nnsamp]={0},nreps[nnsamp]={0};
  int finis;
  int stnindx;

  float apcp[mxyrs],rapcp[mxyrs],rnapcp[mxyrs],runmean[mxyrs];
  float nmodel[mxyrs],rnmodel[mxyrs];
  float dapcp[mxyrs],lpapcp[mxyrs],hpapcp[mxyrs];
  float umean[nnsamp]={0.0},usig[nnsamp]={0.0};
  float mcumean[nnsamp]={0.0},mcusig[nnsamp]={0.0};
  float acor[3]={0.0}, aic[3]={0.0}, phibest[3], errvr[3], z, c, r2;
  float yint,slpct,pcpmean,pcpvrnc,mmean,mvrnc;
  float npvrnc,hpvrnc,lpvrnc; /* np=no-pass hp=hi-pass lp=lo-pass */
  float npmean,hpmean,lpmean;
  float rmin,rmax;

  float dslope,zslope,mslope,mcount;
  float wvrnc=.083;
  /* tc1 = Cutoff period for Lanczos filter, e.g.  
           for min nsamp =6 years assume cutoff frequency of 1 cycle every 12 years, tc1 = 1/12.0
           set = 0.0 for simple detrend as in cpd-11-755-2015 */
  float tc1 = 0.0; 
  int tsamp = 21; /* Box Car Averaging window for plotting purposes, 
                     e.g. Figs. 2a,c-e and Figs 3a-d in cpd-11-755-2015  */ 
  float zthresh = 1.645; /* only write Z stats significant at 90% or better significance level */

  double mwu=0.0,mwz,z0,z1;
  double sig[nmax+1],absig[nmax+1];
  double mslope_mean,mslope_vrnc;

  FILE *ofpntr1, *ofpntr2;
  FILE *ifpntr1, *ifpntr2;

  ifpntr1 = fopen("data.in", "r");

  /* Get Data Record from STDIN... see CALC_ORR.scr  */ 
  nyrs = 0;
  while (fgets(buf, sizeof(buf), ifpntr1) != NULL)
   {  
     sscanf(buf,"%d %f",&year,&apcp[nyrs]);
     if(nyrs == 0)ibegyr=year;
     /* printf("%4d %4d  %11.2f\n",nyrs,year,apcp[nyrs]); optional  */
     nyrs++;
    }
  fclose(ifpntr1);
  iendyr=year;

  /* Read Station Index... In cpd-11-755-2015 stnindx = 1-57 in Fig. 4  */
  std::cin.getline(cbuf, sizeof(cbuf));    
  sscanf(cbuf,"%d ",&stnindx);

  /* calculate mean & vrnc of raw data */
  getacor(apcp, acor, nyrs, &npmean,&npvrnc);

  /* printf("%s\n",model_name); */

  rank(apcp,rapcp,nyrs); /* rank the data */

  calc_runmean(apcp,runmean,nyrs,tsamp); /* calc running mean of data over tsamp year windows  */

  /* Calculate classical MWU parameters for each sample size... */
   for(n = 1; n  <= nnsamp; n++)
     {
      nsamp[n-1]=shrtwin+n-1;
      no[n-1]=nyrs-nsamp[n-1];
      umean[n-1]=.5*no[n-1]*nsamp[n-1]; /*  MWU pop. mean */
      usig[n-1]=sqrt((no[n-1]*nsamp[n-1]*nsamp[n-1] + nsamp[n-1]*no[n-1]*no[n-1]
		      +  nsamp[n-1]*no[n-1])/12.0); /*  MWU st.deviation */
      nreps[n-1]=nyrs/nsamp[n-1];  
     }
     printf("\n"); 

    /* Calculate MC derived U stat Parameters for Station */

    /* first, model time series ... */

    /* Two Options: 1) tc1=0.0: Calculate AR coefficients from detrended data or,
                    2) tc1 != 0.0: from high passed data series...  
       then, calculate lag 1-2-3 autocorrelation values  */ 
     if(tc1 == 0.0)
      {
        detrend(apcp,dapcp,nyrs,&dslope,&yint,&slpct); 
	      getacor(dapcp,acor,nyrs,&pcpmean,&pcpvrnc); 
     /* for(n = 1; n <= 3; n++)printf("ACOR %d %9.5f\n",n,acor[n-1]);
          printf("\n"); */ 
        /* printf("SLOPE = %9.3f\n",slope);  
           printf("YINT  = %9.3f\n",yint); 
           printf("SLPCT  = %9.3f\n\n",slpct); */ 
       }
     else
      {
	     lanfltr(apcp,lpapcp,hpapcp,nyrs,&tc1);    
	     getacor(hpapcp,acor,nyrs,&pcpmean,&pcpvrnc);  
     /* for(n = 1; n <= 3; n++)printf("ACOR %d %9.5f\n",n,acor[n-1]);
          printf("\n"); */ 
       }
      
     /* echo to STDOUT (Optional) */
     for(n = 1; n <= 3; n++)printf("ACOR %d %9.5f\n",n,acor[n-1]); printf("\n");   
     printf("\n");  

     /* Get Best Auto Regressive model of AR(1),AR(2),AR(3) */
     /* Based on Akaike Information Criteria (aic) */
     getar(acor,aic,phibest,nyrs,errvr,pcpmean,pcpvrnc,&ibest,&z); 
     for(n = 1; n <= 3; n++)printf("PHIBEST %d %9.5f\n",n,phibest[n-1]);
     printf("\n"); 
     for(n = 1; n <= 3; n++)printf("AIC %d %9.5f\n",n,aic[n-1]);
     printf("\n"); 
     printf("IBEST = %d\n",ibest);
     printf("\n"); 
     c = sqrt(errvr[ibest-1]/wvrnc); 

    srand(time(NULL));    

    finis = 0;
    for (n = 1 ; n <= nnsamp; n++)ireps[n-1]=0;
    for (n = 1 ; n <= nnsamp; n++)finis += (ireps[n-1] < nr);
 
    mcount=0.0;
    mslope_mean=0.0;
    mslope_vrnc=0.0; 
    while (finis > 0) /* loop unti 50K U stats have been generated for each sample size */
     {

      /* form ibest Markov process */
      getmarkov3(nmodel, nyrs, &mmean, &mvrnc, pcpmean, c, 
              &phibest[0], &phibest[1], &phibest[2]);
     
      gettrend(nmodel,nyrs,&mslope);
      mcount += 1.0;
      mslope_mean += mslope;
      mslope_vrnc += (mslope*mslope);

      /* rank noise model */
      rank(nmodel,rnmodel,nyrs);

      /* Calculate Monte Carlo derived null statistics */
     for (n1 = 1 ; n1 <= nnsamp; n1++) 
      { 
      for (n2 = 1 ; n2 <= nreps[n1-1]; n2++)  
	     { 
	      if(ireps[n1-1] >= nr)break; /*keep going  until ireps =  nr*/
          ireps[n1-1] += 1;
          r2=0.0;
          for (n3 = (n2-1)*nsamp[n1-1]+1 ; n3 <= n2*nsamp[n1-1]; n3++)
            r2 += rnmodel[n3-1];
          mwu = r2 - .5*nsamp[n1-1]*(nsamp[n1-1]+1);
          mcumean[n1-1] += mwu;
          mcusig[n1-1] += mwu*mwu;
         }
      }    
     finis=0; 
     for (n = 1 ; n <= nnsamp; n++)finis += (ireps[n-1] < nr); 
     /* finis set = 0 when 50K null stats generated for every sample size */ 

   } /* while (finis > 0) etc */

    mslope_mean /= mcount;
    mslope_vrnc=(mslope_vrnc-(mcount*(mslope_mean*mslope_mean)))/(mcount-1.0);
    mslope_vrnc=sqrt(mslope_vrnc);
    zslope=(dslope-mslope_mean)/mslope_vrnc;

    ofpntr1 = fopen("series.dat","w");
    for(iyr = 0; iyr <= nyrs-1; iyr++)
        if(tc1 == 0.0) /* write year, data series, data rankings, running mean of data , detrended data series, & noise model series sample to series.dat */ 
         {   
          fprintf(ofpntr1,"%4d %11.2f %11.2f %11.2f %11.2f %11.2f\n",
	          iyr+ibegyr,apcp[iyr],rapcp[iyr],runmean[iyr],dapcp[iyr],nmodel[iyr]);  
          }
         else /* write year, data series, data rankings, running mean of data , low pass & high pass data series, & noise model series sample to series.dat */  
         {
	        fprintf(ofpntr1,"%4d %11.2f %11.2f %11.2f %11.2f %11.2f %11.2f\n",
	          iyr+ibegyr,apcp[iyr],rapcp[iyr],runmean[iyr],lpapcp[iyr],hpapcp[iyr],nmodel[iyr]);  
          }
    fclose(ofpntr1); 

    /* Calculate MC-generated null parameters & record results to STOUT  = calc_orr.log   
       Monte Carlo parameters are recorded as ratio of classical MWU parameters (Eq. 1,2 of cpd-11-755-2015) */
    for (n = 1 ; n <= nnsamp; n++)
       {
	       mcumean[n-1] /= (float)ireps[n-1];
         mcusig[n-1] = sqrt((mcusig[n-1] - (float)ireps[n-1]*mcumean[n-1]*mcumean[n-1])
			   /(float)(ireps[n-1]-1));  
         printf("NSAMP %5d NO %5d NREPS %5d UMEAN %9.2f RATIO %9.2f USIG %9.2f RATIO %9.2f\n", 
	       nsamp[n-1],no[n-1],nreps[n-1],umean[n-1],mcumean[n-1]/umean[n-1], 
	       usig[n-1],mcusig[n-1]/usig[n-1]); 
       }
     printf("\n"); 
 

   /* Calculate running Z stats over all sample sizes and screen for significance */

   /* initialize all holding arrays */
    for (n  = 1 ; n <= nmax; n++)
     {
      yr1[n]=0;yr2[n]=0;
      sig[n]=0.0,absig[n]=0.0;
      yrbeg[n]=0;yrend[n]=0;
      index[n]=0;
     }
     nsig=0;

   /* for each sampling size, write all significant z stats and
      time windows to holding arrays */

   for (n1 = 1 ; n1 <= nnsamp; n1++)  
     {

      for (n2 = ibegyr ; n2 <= iendyr-nsamp[n1-1]+1; n2++)  
	     { 
          r2=0.0;
          iyrb=n2;
          iyre=n2+nsamp[n1-1]-1;
          for (n3 = iyrb; n3 <= iyre; n3++)
            r2 += rapcp[n3-ibegyr];
          mwu = r2 - .5*nsamp[n1-1]*(nsamp[n1-1]+1);

          /* mwz output options */

          /*  Z stats via classical stats */
          /* mwz = z0 = (mwu - umean[n1-1])/usig[n1-1];  */

          /* or Z stats via Monte Carlo stats */
          /* mwz = z1 = (mwu - mcumean[n1-1])/mcusig[n1-1]; */

          /* Z stats determined via  fmin(fabs(z0),fabs(z1)) 
           See discussion on lines 1-12 page 11 of cpd-11-755-2015  */
          z0 = (mwu - umean[n1-1])/usig[n1-1];  
          z1 = (mwu - mcumean[n1-1])/mcusig[n1-1];
          mwz = fmin(fabs(z0),fabs(z1));  
          if(mwz == fabs(z0))mwz=z0;
          if(mwz == fabs(z1))mwz=z1;
 	   
	  if(fabs(mwz) >=  zthresh)
	    {nsig += 1;
	     absig[nsig]=fabs(mwz);
             sig[nsig]=mwz; 
             yrbeg[nsig]=iyrb;
             yrend[nsig]=iyre;
             index[nsig]=nsig;} 
         }
     
     }  /* for (n1 = 1 ; n1 <= nnsamp; n1++) */ 

      printf("\n");

   /* sort all significant results by abs(Z) */
   
   /* parallel sort absig and index by absig */
   bubble1(absig,index,&nsig); 
   
   ofpntr1 = fopen("calc_orr.dat","w");
   ofpntr2 = fopen("calc_orr.dat.all","w");

   /* write highest ranked results */
   fprintf(ofpntr1,"%3d %4d %4d %7.4f %7.4f\n",
      stnindx,yrbeg[index[1]],yrend[index[1]],sig[index[1]],absig[1]);
   fprintf(ofpntr2,"%3d %4d %4d %7.4f %7.4f\n",
      stnindx,yrbeg[index[1]],yrend[index[1]],sig[index[1]],absig[1]);
  
   nwrote=1;
   yr1[nwrote]=yrbeg[index[1]];
   yr2[nwrote]=yrend[index[1]]; 

   /* of the remaining significant values in absig, write only those
         that do not overlap in time with previously written values to calc_orr.dat 
         write all significant results to calc_orr.dat.all */

    for(int i = 2; i <= nsig; i++)
	     {            
             pass=1; 
             /* write to calc_orr.dat.al */
             fprintf(ofpntr2,"%3d %4d %4d %7.4f %7.4f\n",
	             stnindx,yrbeg[index[i]],yrend[index[i]],sig[index[i]],absig[i]);
             for(int j = 1; j <= nwrote; j++)
	            {
                 if(yrbeg[index[i]] >= yr1[j] && yrbeg[index[i]] <= yr2[j]){pass=0;break;}
                 if(yrend[index[i]] >= yr1[j] && yrend[index[i]] <= yr2[j]){pass=0;break;}
            		 if(yrbeg[index[i]] <= yr1[j] && yrend[index[i]] >= yr2[j]){pass=0;break;} 
               }
	           if(pass==1)
	            {
                fprintf(ofpntr1,"%3d %4d %4d %7.4f %7.4f\n",
                 stnindx,yrbeg[index[i]],yrend[index[i]],sig[index[i]],absig[i]);
                 nwrote+=1;
	               yr1[nwrote]=yrbeg[index[i]];
                 yr2[nwrote]=yrend[index[i]];
	             } 
                           
	  } /* for(int i = 2; i <= nsig; i++) */   

   fclose(ofpntr1);
   fclose(ofpntr2);
   exit:
   return 0;
}
/*----------------------------- FUNCTIONS --------------------------------------*/


void bubble1(double array[mxyrs], int index[mxyrs] , int* nsig)
{
  int pass,count;
  void dswap (double *,double *);
  void intswap (int *,int *);
 
  for (pass = 1; pass <= *nsig -1 ; pass++)
    for (count=1 ; count <= *nsig - 1 - (pass-1) ; count++)
      if(array[count] < array[count+1])
        {
         dswap(&array[count],&array[count+1]);
         intswap(&index[count],&index[count+1]);
        }
}

void dswap (double *ptr1, double *ptr2)
{
  double temp;
  temp = *ptr1;
  *ptr1 = *ptr2;
  *ptr2 = temp;
}

void intswap (int *ptr1, int *ptr2)
{
  int temp;
  temp = *ptr1;
  *ptr1 = *ptr2;
  *ptr2 = temp;
}

void calc_runmean(float aflo[mxyrs], float runmean[mxyrs], int nyrs, int tsamp)
{
  float null=-999.99;
  int iyr, iwin, i;
  iwin=(tsamp-1)/2;
  for (iyr = iwin; iyr <= nyrs-iwin-1; iyr++)
    {
     for (i = iyr-iwin; i <= iyr+iwin; i++){runmean[iyr] += aflo[i];}
     runmean[iyr] = runmean[iyr]/tsamp;
    }
  for (iyr = 0; iyr <= iwin-1; iyr++){runmean[iyr]=null;}
  for (iyr = nyrs-iwin; iyr <= nyrs-1; iyr++){runmean[iyr]=null;}
}


void rank(float nmodel[mxyrs], float rnmodel[mxyrs], int nyrs)
{
  float time[mxyrs],work[mxyrs],null=-99.9,del=.009;
  int iyr;

  void bubble2 (float nmodel[mxyrs],float rnmodel[mxyrs], int nyrs);

 /* parallel rank aflow and time by aflow */

   for (iyr = 0; iyr <= nyrs-1; iyr++)
    {time[iyr] = iyr + 1.;
     work[iyr] = nmodel[iyr];
     rnmodel[iyr] = null;}

   bubble2(nmodel,time,nyrs); /* rank nmodel and time by value */

   /* assign raflow based on position in rank sequence */
   /* assign tie if rabs(aflow[iyr],aflow[iyr+1]) < del */

   for (iyr = 0; iyr <= nyrs-2; iyr++)
    {
     if(rnmodel[iyr] == null) /* then not ranked yet */
     {
      if(rabs(nmodel[iyr],nmodel[iyr+1]) >  del) 
       {rnmodel[iyr]= iyr + 1.;}
      else
       {rnmodel[iyr]= iyr + 1.5;
        rnmodel[iyr+1]= iyr + 1.5;}
      }
    }
    if(rnmodel[nyrs-1] == null)rnmodel[nyrs-1]=nyrs;

    /* parallel sort time and raflow by time, which  resorts ranks back */ 
    /*into their proper sequence in time */

    bubble2(time,rnmodel,nyrs);

    for (iyr = 0; iyr <= nyrs-1; iyr++)nmodel[iyr] = work[iyr];

}

void bubble2(float array[mxyrs], float rank[mxyrs], int nyrs)
{
  int pass,count;
  void swap (float *,float *);
 
  for (pass = 1; pass <= nyrs -1 ; pass++)
    for (count=0 ; count <= nyrs -2; count++)
      if(array[count] > array[count+1])
        {
         swap(&array[count],&array[count+1]);
         swap(&rank[count],&rank[count+1]);
        }
}

void swap (float *ptr1, float *ptr2)
{
  float temp;
  temp = *ptr1;
  *ptr1 = *ptr2;
  *ptr2 = temp;
}

float rabs(float a ,float b)
{
  if (a > b)return a-b;
  else return b-a;
}

void getacor(float dser[mxyrs], float acor[maxlag], int nyrs, float* flmean, float* flvrnc)
{  
  float count;
  int n,k;

  *flmean=0.0;
  *flvrnc=0.0;
  for(n = 0; n <= nyrs-1; n++)
    { *flmean += dser[n]; *flvrnc += dser[n]*dser[n]; }
  *flmean /= (float)nyrs;
  *flvrnc -= ((float)nyrs * *flmean * *flmean);
  *flvrnc /= (float)(nyrs-1);

  for(k = 1 ; k <= maxlag; k++)
    {
     count = 0.0;  
     for(int n = 1 ; n <= nyrs-k; n++)
	count += ((dser[n-1] - *flmean)*(dser[n+k-1] - *flmean)); 
     acor[k-1] = (count/(float)(nyrs-k-1))/(*flvrnc); 
     /* printf("LAG = %d ACOR = %f\n",k,acor[k-1]); */  
     } 
}

  
void detrend(float ser[mxyrs],float dser[mxyrs], int nyrs, float* slope, float* yint, float* slpct)
{
  float x=0.0,x2=0.0,xy=0.0,y=0.0;
  float ymean,del;
  int n;

  for(n = 1; n <= nyrs; n++)
    {
      x += (float)n;
      x2 += ((float)n*(float)n);
      xy += (float)n * ser[n-1];
      y += ser[n-1];
    }
  
  del = (nyrs*x2) - (x*x);
  *yint = (x2*y - x*xy)/del;
  *slope = (nyrs*xy - x*y)/del;
  ymean = y/((float)nyrs); 

  for(n = 1; n <= nyrs; n++)
     dser[n-1] = (ser[n-1] - (*slope*(float)n + *yint)) + ymean; 

  /* slope pct = rise divided by mean */
  *slpct = 100.*(*slope*(float)nyrs)/(ymean); 

}

void gettrend(float ser[mxyrs], int nyrs, float* slope)
{
  float x=0.0,x2=0.0,xy=0.0,y=0.0;
  float ymean,del;
  int n;

  for(n = 1; n <= nyrs; n++)
    {
      x += (float)n;
      x2 += ((float)n*(float)n);
      xy += (float)n * ser[n-1];
      y += ser[n-1];
    }
  
  del = (nyrs*x2) - (x*x);
  *slope = (nyrs*xy - x*y)/del; 

}

void getar(float acor[maxlag], float aic[maxlag], float phibest[maxlag], int nyrs,
           float errvr[maxlag],float flmean, float flvrnc, int* ibest, float* z)
{           
  float phi[3][3];
  float a,b,c,d;
  float aicmin;
  int n;
  double c2;
 
  /* See Box and Jenkins pages 54-55 */
  /* temporary variables for Yule-Walker solutions */
  a = 1.0 - acor[0]*acor[0];
  b = 1.0 - acor[1]*acor[1];
  c = acor[0]*(1.0 - acor[1]);
  d = acor[1] - (acor[0]*acor[0]);

  /* AR1 coefficients... */
  phi[0][0]=acor[0];
  phi[0][1]=0.0;
  phi[0][2]=0.0;

  /* AR2... */
  phi[1][0]=c/a;
  phi[1][1]=d/a;
  phi[1][2]=0.0;

  /* AR3... */
  phi[2][2] = (a*acor[2] - c*d - a*acor[0]*acor[1])/(b*a - c*c);
  phi[2][1] = (acor[1] - acor[0]*acor[0] -  c*phi[2][2])/a;
  phi[2][0] = acor[0]*(1. - phi[2][1]) - acor[1]*phi[2][2];

  /* Error Variances as in Wilks Ed. 1 pages 306- 309 */ 
  errvr[0] = flvrnc*(1.-(phi[0][0]*phi[0][0]))*((float)nyrs - 1.)/
                        ((float)nyrs - 2.);
  errvr[1] = errvr[0] * (1. - phi[1][1]*phi[1][1]);
  errvr[2] = errvr[1] * (1. - phi[2][2]*phi[2][2]);

  c2 = (1.0 - (phi[0][0]*phi[0][0]))/(float)nyrs;
  *z = phi[0][0]/sqrt(c2);
      
  /* Select best model via Akaike Information Criteria */
  aicmin = FLT_MAX;
  for (n = 1; n <= maxlag; n++)
   {
     c2 = (float)nyrs/(float)(nyrs - n - 1) * (double)errvr[n-1];
     aic[n-1] = (float)nyrs * log(c2) + 2.*(float)n + 2.;
     if (aic[n-1] < aicmin) {*ibest = n; aicmin = aic[n-1];}
   }

   for (n = 1; n <= maxlag; n++)
     printf(" 'MODEL ORDER'  %3d 'AIC'' %9.2f 'REL ERROR'' %7.2f\n",
     n,aic[n-1],100.*errvr[n-1]/flvrnc);  
    printf("\n");
     
   
  /* Select lowest AIC as best */
  for (n =1; n <=maxlag; n++)
    phibest[n-1] = phi[*ibest-1][n-1];

}


void getmarkov3(float nmodel[mxyrs], int nyrs, float* mmean, float* mvrnc, float flmean, 
                float c, float* phi1, float* phi2, float* phi3)
{
  float work[nyrs+3],rmax;
  int n,ic;  

  rmax=(float)RAND_MAX;

  work[0] = flmean*(1.- *phi1 - *phi2 - *phi3) + c*((float)rand()/rmax - .5);
  work[1] = *phi1*work[0] 
                    + flmean*(1.-*phi1 - *phi2 - *phi3) + c*((float)rand()/rmax - .5);
  work[2] = *phi1*work[1] + *phi2*work[0] 
                    + flmean*(1.-*phi1 - *phi2 - *phi3) + c*((float)rand()/rmax - .5);

  *mmean=0.0;
  *mvrnc=0.0;
  ic = 0;
  for(n = 4; n <= nyrs+3; n++)
   {
   work[n-1]= *phi1*work[n-2] + *phi2*work[n-3] + *phi3*work[n-4]
                    + flmean*(1.- *phi1 - *phi2 - *phi3) + c*((float)rand()/rmax - .5);
   nmodel[n-4]=work[n-1];
   *mmean += work[n-1];
   *mvrnc += work[n-1]*work[n-1];
   }
  *mmean /= (float)nyrs;
  *mvrnc = (*mvrnc - (float)nyrs*(*mmean)*(*mmean))/(float)(nyrs-1); 

} 


void lanfltr(float ser[mxyrs],float lpser[mxyrs], float hpser[mxyrs], int nyrs, float* tc1)
{

  void lan(float *, int *, float w[301], float r[151], float freq[151], int * );

  float fc1;
  int slength=nyrs;
  int slength2=2*slength;
  float work[3*slength + 1];
  float w[301],r[151],freq[151];
  float mean;
  int nwt=101,n,iflt;
  int kop=1;

  iflt=(nwt-1)/2;
  fc1=1.0/ *tc1;

  /* Load Data into the central 1/3 of the work array */
  for(int i = slength+1; i <= slength2; i++)work[i]=ser[i-(slength+1)];

  /* Reflect data into pre- and post-data period... */
  mean=0.0;
  for(int i = 1; i <= slength; i++)
  {
    mean += ser[i-1];
    work[i]=work[slength2-(i-1)];
    work[slength2+i]=work[slength2-(i-1)];
  }
  mean /= (float)slength;

  /* calculate low pass filter weightings */
  lan(&fc1, &iflt, w, r, freq, &kop);

  /*  Filter away... */
  for(int iyr = 0; iyr <= slength-1; iyr++)lpser[iyr]=0.0;

  for(int iyr = slength+1; iyr <= slength2; iyr++)
  {
    lpser[iyr-(slength+1)]=w[iflt+1]*work[iyr];
    for(int i = 1; i <= iflt; i++)lpser[iyr-(slength+1)]+=w[iflt+i+1]*work[iyr+i];
    for(int i = 1; i <= iflt; i++)lpser[iyr-(slength+1)]+=w[iflt+i+1]*work[iyr-i];
  }
  /* high pass output = (data - low pass) + mean */ 
  for(int iyr = 0; iyr <= slength-1; iyr++)
   hpser[iyr]=(ser[iyr]-lpser[iyr]) + mean;

}

void lan(float* fc1, int* iflt, float w[301], float r[151], float freq[151], int* kop)
{
  float pi=3.14159265;
  float pi2,t,s,x,y,f,z,w0;
  float a[301];
  int nw,l;

  pi2=2.0*pi;
  nw = 2 * *iflt + 1;
  t = pi2 * *fc1;
  s=0.0;

   for(int k = 1; k <= *iflt; k++)
    {
      x=pi*(float)k;
      y=sin(t * (float)k)/x;
      a[k]=y*sin(x/((float)*iflt))/(x/((float)*iflt));
      s += (2.0 * a[k]);
     } 

  w0 = 1.0 - s;
  l = 0;

  for(int i = *iflt; i >= 1; i--)
    {
      l += 1;
      w[l]=a[i];
    }

  w[*iflt + 1] = w0;

  for(int i = 1; i <= *iflt; i++)
    {
      w[i + *iflt + 1] = a[i];
    }

  for(int i = 1; i <= 150; i++)
    {
      f = 0.0 + (float)(i-1) * 0.00333;
      freq[i] = f;
      z=0.0;
      for(int k = 1; k >= *iflt; k++)z += a[k]*cos(pi2*f*(float)k);
      r[i] = (w0 + 2.0*z)* (w0 + 2.0*z);
      if(*kop == 4)r[i]=1.0-r[i];
    }

}

float array_min(float array[nnsamp])
{
  int i;
  float min = FLT_MAX;
  for (i = 0; i <= nnsamp-1; i++)
    {if(array[i] < min)min=array[i];}
  return min;
}

float array_max(float array[nnsamp])
{
  int i;
  float max = FLT_MIN;
  for (i = 0; i <= nnsamp-1; i++)
    {if(array[i] > max)max=array[i];}
  return max;
}





